from typing import List, Dict, Union, Optional, Set
from uuid import UUID
from pyarrow import csv, json, Table

from datalogue.clients._http import (
    _HttpClient,
    HttpMethod,
    MultiPartData,
    MultiPartFile,
)
from datalogue.models.schema import AbstractDataSchema, Schema, SchemaNode
from datalogue.models.datastore import (
    Datastore,
    _datastore_from_payload,
    FileFormat,
    Cell,
    _cell_from_payload,
    NodeClassification,
    ReclassificationBehavior,
    RollupStrategy,
    WeightedAverage,
)
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list, ResponseStream
from datalogue.clients._processed_sources import _ProcessedSourcesClient


class _MetadataClient:
    """
    Client to interact with metadata
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def add(
        self, datastore_id: Union[str, UUID], field: List[str], metadata: Dict[str, str]
    ) -> Union[DtlError, SchemaNode]:
        """
        Add key:value pairs of metadata to a field

        :param datastore_id: id of the datastore to which to add metadata
        :param field: path to the field to which to add metadata; note: the specified metadata will be added to all fields thus named
        :param metadata: a dictionary of metadata, in key:value format, to be added to the afore-specified field; e.g. {'description':'family names of customers', 'privacy':'private'} will add two pairs of metadata
        :return: the updated SchemaNode if successful or DtlError if failed
        """
        payload = {"fieldPath": field, "metadata": metadata}

        rsp = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/metadata",
            HttpMethod.PUT,
            payload,
        )

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return SchemaNode._schema_node_from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def remove(
        self, datastore_id: Union[str, UUID], field: List[str], key: str
    ) -> Union[DtlError, SchemaNode]:
        """
        Remove a specified metadata key:value pair from a field

        :param datastore_id: id of the datastore from which to remove metadata
        :param field: path to the field from which to remove metadata; note: the specified metadata will be removed from all fields thus named
        :param key: the key of the metadata pair you wish to remove
        :return: the updated SchemaNode if successful or DtlError if failed
        """
        payload = {"fieldPath": field, "key": key}

        rsp = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/metadata",
            HttpMethod.DELETE,
            payload,
        )

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return SchemaNode._schema_node_from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def replace_value(
        self, datastore_id: Union[str, UUID], field: List[str], key: str, new_value: str
    ) -> Union[DtlError, SchemaNode]:
        """
        Replace the value of a specified metadata key:value pair

        :param datastore_id: id of the datastore of which to replace metadata value
        :param field: path to the field of which to replace metadata value
        :param key: the key of the metadata pair whose value should be replaced
        :new_value: the new value of the afore-specified key
        :return: the updated SchemaNode if successful or DtlError if failed
        """
        metadata = {}
        metadata[key] = new_value
        payload = {"fieldPath": field, "metadata": metadata}

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/metadata/replace",
            HttpMethod.POST,
            body=payload,
        )
        if isinstance(res, DtlError):
            return res

        return SchemaNode._schema_node_from_payload(res)

    def clear(
        self, datastore_id: Union[str, UUID], field: List[str]
    ) -> Union[DtlError, SchemaNode]:
        """
        Clear all metadata from a field

        :param datastore_id: id of the datastore from which to clear metadata
        :param field: path to the field from which to clear metadata
        :return: the updated SchemaNode if successful or DtlError if failed

        """
        payload = {"fieldPath": field}

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/metadata/clear",
            HttpMethod.POST,
            body=payload,
        )
        if isinstance(res, DtlError):
            return res

        return SchemaNode._schema_node_from_payload(res)


class DatastoreClient:
    """
    Client to interact with the Datastores objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"
        self.metadata = _MetadataClient(http_client)
        self.processed_sources = _ProcessedSourcesClient(self.http_client)

    def create(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Creates a data source

        :param datastore: data source to be created
        :return: string with error message if failed, the data source otherwise
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores", HttpMethod.POST, datastore._as_payload()
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def get(self, datastore_id: UUID) -> Union[DtlError, Datastore]:
        """
        From the provided id, get the corresponding Datastore

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id), HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return res

        datastore_resp = _datastore_from_payload(res)
        if isinstance(datastore_resp, DtlError):
            return datastore_resp

        return self.__load_datastore_with_class_info(datastore_resp)

    def get_by_credential_id(
        self, credential_id: UUID, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Datastore]]:
        """
        From the provided credential id, get corresponding Datastores
        :param credential_id: id of the datastore to be retrieved
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all datastores for given credential_id or an error message as a string
        """
        payload = {
            "queries": {},
            "from": (page - 1) * item_per_page,
            "size": item_per_page,
            "type": "datastore",
        }

        res = self.http_client.make_authed_request(
            self.service_uri + f"/credentials/{str(credential_id)}/datastores/search",
            HttpMethod.POST,
            payload,
        )

        if isinstance(res, DtlError):
            return res

        datastore_list = _parse_list(_datastore_from_payload)(res["results"])
        if isinstance(datastore_list, DtlError):
            return datastore_list

        datastore_list_with_class_info = []
        for datastore_resp in datastore_list:
            datastore_resp_with_class_info = self.__load_datastore_with_class_info(
                datastore_resp
            )
            if isinstance(datastore_resp_with_class_info, DtlError):
                return datastore_resp_with_class_info
            datastore_list_with_class_info.append(datastore_resp_with_class_info)
        return datastore_list_with_class_info

    def update(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Updates the backend with the new status of the existing datastore

        :param datastore: to be persisted
        :return:
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore.id),
            HttpMethod.PUT,
            datastore._as_payload(),
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def list(
        self, by_tag_name: Optional[str] = None, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Datastore]]:
        """
        List all the datastores that are saved

        :param by_tag_name: optionally filter your list by only those datastores tagged by the specified tag
        :param page: page to be retrieved (ignored)
        :param item_per_page: number of items to be put in a page (ignored)
        :return: Returns a List of all the available datastores or an error message as a string
        """
        params = {
            "page": page,
            "size": item_per_page,
        }
        if isinstance(by_tag_name, str):
            params["tag_name"] = by_tag_name

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores", HttpMethod.GET, params=params
        )

        if isinstance(res, DtlError):
            return res

        datastore_list = _parse_list(_datastore_from_payload)(res)
        if isinstance(datastore_list, DtlError):
            return datastore_list

        datastore_list_with_class_info = []
        for datastore_resp in datastore_list:
            datastore_resp_with_class_info = self.__load_datastore_with_class_info(
                datastore_resp
            )
            if isinstance(datastore_resp_with_class_info, DtlError):
                return datastore_resp_with_class_info
            datastore_list_with_class_info.append(datastore_resp_with_class_info)
        return datastore_list_with_class_info

    def search(
        self,
        by_tag_name: Optional[str] = None,
        by_name: Optional[List[str]] = None,
        by_field_name: Optional[List[str]] = None,
        by_field_class: Optional[List[str]] = None,
        page: int = 1,
        item_per_page: int = 25,
    ) -> Union[DtlError, List[Datastore]]:
        """
        Search among all backend Datastores, using any numbers of queries as the search criteria
         Multiple items within a query will be applied with OR logic — multiple queries will be applied with AND logic

         The wildcard operators: `?` and `*` can be used as wildcards,
         matching one and any number of additional characters, respectively

        :param by_tag_name: a tag applied to the datastores, specified by the full tag name
        :param by_name: a list of datastore names
        :param by_field_name: a list of column labels
        :param by_field_class: a list of classes that columns are classified
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available datastores or an error message as a string
        """

        queries = {}

        if isinstance(by_name, list):
            queries["name"] = by_name

        if isinstance(by_field_name, list):
            queries["schema.nodes.label"] = by_field_name

        if isinstance(by_field_class, list):
            queries["classes.name"] = by_field_class

        if len(queries) > 0:
            payload = {
                "queries": queries,
                "from": (page - 1) * item_per_page,
                "size": item_per_page,
                "type": "datastore",
            }
            if isinstance(by_tag_name, str):
                payload["tag"] = by_tag_name

            res = self.http_client.make_authed_request(
                self.service_uri + "/v2/search", HttpMethod.POST, payload
            )
            if isinstance(res, DtlError):
                return res

            return _parse_list(_datastore_from_payload)(res["results"])
        else:
            # If there is not query, we will bring everything
            payload = {
                "query": "*",
                "page": page,
                "size": item_per_page,
                "type": "datastore",
            }
            if isinstance(by_tag_name, str):
                payload["tag"] = by_tag_name

            res = self.http_client.make_authed_request(
                self.service_uri + "/search", HttpMethod.POST, payload
            )
            return _parse_list(_datastore_from_payload)(res["results"])

    def get_schema(self, datastore_id: UUID) -> Union[DtlError, AbstractDataSchema]:
        """
        From the provided id, get the corresponding Datastore schema

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/schema",
            HttpMethod.GET,
        )
        if isinstance(res, DtlError):
            return res

        schema = Schema._from_payload(res).data_schema
        if isinstance(schema, DtlError):
            return schema

        schema = self.__load_ads_with_class_info(schema)
        return schema

    def __load_datastore_with_class_info(
        self, datastore: Datastore
    ) -> Union[DtlError, Datastore]:
        if datastore.ads is None and (
            datastore.schema_nodes is None or (len(datastore.schema_nodes) == 0)
        ):
            return datastore

        if datastore.ads is not None or datastore.schema_nodes is not None:
            # 1. Get schema nodes from either Datastore or from ADS
            schema_nodes = []
            if datastore.schema_nodes is not None:
                schema_nodes = datastore.schema_nodes
            elif datastore.ads is not None:
                schema_nodes = list(
                    map(lambda node_def: node_def.node, datastore.ads.get_nodes())
                )

            # 2. Load the dictionary for set of class_id
            class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)
            if isinstance(class_id_path_dict, DtlError):
                return class_id_path_dict

            # 3. Update ADS and List of SchemaNode respectively
            if datastore.ads is not None:
                datastore.ads = self.__fill_ads(datastore.ads, class_id_path_dict)

            if datastore.schema_nodes is not None:
                datastore.schema_nodes = self.__fill_schema_nodes(
                    datastore.schema_nodes, class_id_path_dict
                )

            if datastore.schema_classes is not None:
                for schema_class in datastore.schema_classes:
                    class_path = class_id_path_dict.get(schema_class.class_id)
                    if class_path is None:
                        # There is no class or user does not have Read permissions
                        schema_class.class_name = "Permission denied"
                        schema_class.class_path = ["Permission denied"]
                    elif len(class_path) > 0:
                        schema_class.class_name = class_path[-1]
                        schema_class.class_path = class_path
                    else:
                        schema_class.class_path = class_path

        return datastore

    def __load_ads_with_class_info(
        self, schema: AbstractDataSchema
    ) -> Union[DtlError, AbstractDataSchema]:
        schema_nodes = list(map(lambda node_def: node_def.node, schema.get_nodes()))
        class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)

        if isinstance(class_id_path_dict, DtlError):
            return class_id_path_dict

        return self.__fill_ads(schema, class_id_path_dict)

    def __fill_ads(
        self, schema: AbstractDataSchema, class_id_path_dict: Dict[UUID, List[str]]
    ) -> Union[DtlError, AbstractDataSchema]:
        schema_nodes = list(map(lambda node_def: node_def.node, schema.get_nodes()))
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_path = class_id_path_dict.get(classification.class_id)
                if class_path is None:
                    # There is no class or user does not have Read permissions
                    classification.class_name = "Permission denied"
                    classification.class_path = ["Permission denied"]
                elif len(class_path) > 0:
                    classification.class_name = class_path[-1]
                    classification.class_path = class_path
                else:
                    classification.class_path = class_path

        return schema

    def __load_schema_nodes_with_class_info(
        self, schema_nodes: List[SchemaNode]
    ) -> Union[DtlError, List[SchemaNode]]:
        class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)
        return self.__fill_schema_nodes(schema_nodes, class_id_path_dict)

    def __fill_schema_nodes(
        self, schema_nodes: List[SchemaNode], class_id_path_dict: Dict[UUID, List[str]]
    ) -> Union[DtlError, List[SchemaNode]]:
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_path = class_id_path_dict.get(classification.class_id)
                if class_path is None:
                    # There is no class or user does not have Read permissions
                    classification.class_name = "Permission denied"
                    classification.class_path = ["Permission denied"]
                elif len(class_path) > 0:
                    classification.class_name = class_path[-1]
                    classification.class_path = class_path
                else:
                    classification.class_path = class_path
        return schema_nodes

    def __load_paths_of_class_ids(
        self, schema_nodes: List[SchemaNode]
    ) -> Union[DtlError, Dict[str, List[str]]]:
        class_id_set = set([])
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_id_set.add(classification.class_id)

        res = []
        if len(class_id_set) > 0:
            res = self.http_client.make_authed_request(
                f"/yggy/entities?page=1&size={len(class_id_set)}",
                HttpMethod.PUT,
                list(map(lambda x: str(x), class_id_set)),
            )

            if isinstance(res, DtlError):
                return res

        class_id_path_dict = {}
        for classInfo in res:
            class_id = UUID(classInfo["id"])
            class_path = classInfo["path"]
            class_id_path_dict[class_id] = class_path

        return class_id_path_dict

    def print(
        self, datastore_id: UUID, full_path: bool = False
    ) -> Union[DtlError, None]:
        """
        TEMPORARILY DISABLED.
        From the provided id, get the corresponding Datastore schema

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """
        print("Feature currently disabled, please contact Datalogue.")
        # res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id) + "/schema", HttpMethod.GET)
        # if isinstance(res, DtlError):
        #     if res.message.__contains__("Route not found"):
        #         res.message = "Invalid ID, please verify that your data store exists with dtl.datastore.list()"
        #     return res
        # schema = Schema._from_payload(res)
        # schema.print(full_path)

    def delete(self, datastore_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Datastore

        :param datastore_id: id of the datastore to be deleted
        :return: true if successful, false otherwise
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id), HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def download(
        self,
        datastore_id: UUID,
        file_format: FileFormat,
        path: str,
        limit: Optional[int] = None,
    ) -> Union[DtlError, str]:
        """
        Downloads the resource in a file

        :param datastore_id: id of the resource to download
        :param file_format: File format of the downloaded resource
        :param path: location to save the file
        :param limit: limit the number of elements to be downloaded
        :return: An error or the path to the downloaded file
        """

        params = {
            "fileFormat": file_format.value,
        }

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET,
            params=params,
            stream=True,
        )

        if isinstance(res, DtlError):
            return res

        with open(path, "wb") as f:
            for chunk in res.iter_content(chunk_size=1024 * 32):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return path

    def load_arrow_table(
        self,
        datastore_id: UUID,
        file_format: FileFormat = FileFormat.Csv,
        limit: Optional[int] = None,
        batch_size: int = 1024,
    ) -> Union[DtlError, Table]:
        """
        Loads the resource in an in memory arrow buffer.
        This can be used either with Nvidia RAPIDS or pandas

        pandas:
        df = dtl.client.datastore.load_arrow_table(datastore_id).to_pandas()

        :param datastore_id: if of the resource to load
        :param file_format: can be FileFormat.Csv or FileFormat.Json. Json is useful if the data has commas in it, which will throw an error using csv.
        :param limit: elements to be loaded in the frame
        :param batch_size: size of the batch to be retrieved at a time from the collection
        :return:
        """

        params = {"fileFormat": file_format.value}

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET,
            params=params,
            stream=True,
        )

        if isinstance(res, DtlError):
            return res

        stream = ResponseStream(res.iter_content(batch_size))

        if file_format.value == "Json":
            return json.read_json(stream)
        elif file_format.value == "Csv":
            return csv.read_csv(stream)
        else:
            return DtlError(
                f"{file_format.value} is not supported for this method. Please use csv or json instead."
            )

    def upload(
        self, datastore_id: UUID, file_format: FileFormat, path: str, params: dict = {}
    ) -> Union[DtlError, None]:
        """
        Enable to upload a file to the specified resource through the platform

        :param datastore_id: id of the resource to upload the data through
        :param file_format: format of the file to be uploaded
        :param path: location of the file to upload on disk
        :return:
        """

        with open(path, "rb") as f:
            files = [MultiPartFile("content", f)]

            data = [
                MultiPartData("file-format", file_format.value),
                MultiPartData("params", "{}"),
            ]

            res = self.http_client.post_multipart_data(
                self.service_uri + "/datastores/" + str(datastore_id) + "/upload",
                files,
                data,
            )

            if isinstance(res, DtlError):
                return res

    def classify_fields(
        self,
        datastore_id: UUID,
        classifier_id: Union[str, UUID],
        target_fields: Optional[List[List[str]]] = None,
        sample_size: int = 100,
        rollup_strategy: RollupStrategy = WeightedAverage(),
        reclassification_behavior: ReclassificationBehavior = ReclassificationBehavior.ReplaceAutomatic,
    ) -> Union[DtlError, UUID]:
        """
        Classify the fields of a datastore. Classifications will be found within the ‘classifications’ property of
         each SchemaNode of the datastore.

        :param datastore_id: the id of the datastore to classify the fields of
        :param classifier_id: the classifier to apply classes to the data points, which will be used to calculate the
         classes of the fields
        :param rollup_strategy: describes the intended behavior for classification of fields.
            It can be either ClassThreshold or WeightedAverage which is the default value.
        :param reclassification_behavior: describes the intended behavior if this operation reclassifies a
            field--defaults to replacing only automatic classifications from the specified classifier
        :return: the job id of the pipeline classifying the datastore fields
        """

        if isinstance(reclassification_behavior, str):
            return ReclassificationBehavior.parse_error(reclassification_behavior)

        params = {
            "classifier": str(classifier_id),
            "reclassification-behavior": str(reclassification_behavior.value),
        }

        payload = {}
        if target_fields is not None:
            payload["targetFields"] = target_fields
        if sample_size is not None:
            payload["sampleSize"] = sample_size

        payload["rollupStrategy"] = rollup_strategy._as_payload()

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + f"/classify",
            HttpMethod.POST,
            params=params,
            body=payload,
        )

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return UUID(res.json()["jobId"])
        else:
            return DtlError(res.text)

    def add_classifications(
        self, datastore_id: UUID, classifications: List[NodeClassification]
    ) -> Union[DtlError, List[SchemaNode]]:
        """
        Apply classes to specified fields (append, not override)

        :param datastore_id: the id of the datastore to add classes to
        :param classifications: a list of NodeClassifications, representing the fields to add classes to, and the classes to add
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri
            + "/datastores/"
            + str(datastore_id)
            + "/classifications/add",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)),
        )

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)

    def remove_classifications(
        self, datastore_id: UUID, classifications: List[NodeClassification]
    ) -> Union[DtlError, List[SchemaNode]]:
        """
        Remove classes from specified fields.

        :param datastore_id: the id of the datastore to remove classes from
        :param classifications: a list of NodeClassifications, representing the fields from which classes should be removed, and the classes  to remove from those fields
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri
            + "/datastores/"
            + str(datastore_id)
            + "/classifications/remove",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)),
        )

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)

    def override_classifications(
        self, datastore_id: UUID, classifications: List[NodeClassification]
    ) -> Union[DtlError, List[SchemaNode]]:
        """
        Remove all classes from specified fields, and then apply new classes to those fields.

        :param datastore_id: the id of the datastore whose fields should be overridden
        :param classifications: a list of NodeClassifications, representing the fields that should be cleared and the new classes for  those fields
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri
            + "/datastores/"
            + str(datastore_id)
            + "/classifications/replace",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)),
        )

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)

    def resample(self, datastore_ids: List[UUID]) -> Union[DtlError, List[Datastore]]:
        """
        Trigger resampling jobs for a given list of datastore ids.

        :param datastore_ids: ids of the datastores to refresh schema and samples
        :return: datastore payloads
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/schema-sample-jobs",
            HttpMethod.POST,
            body={"ids": datastore_ids},
        )
        if isinstance(res, DtlError):
            return res
        datastore_list = _parse_list(_datastore_from_payload)(res)

        return datastore_list



import abc
from typing import Optional, Union, Dict, List
from datalogue.models.datastore import CredentialType, ChangeDataSource, FileFormat
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models.s3encryption import Encryption
from datalogue.dtl_utils import SerializableStringEnum
from uuid import UUID


class CredentialsReference:
    def __init__(self, ref_id: UUID, name: Optional[str], ref_type: CredentialType):
        self.id = ref_id
        self.name = name
        self.type = ref_type

    def __eq__(self, other: "CredentialsReference"):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.name == other.name and self.type == other.type
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}(id= {self.id}, name= {self.name!r}, type= {self.type!r})"


def _credentials_ref_from_payload(json: dict) -> Union[DtlError, CredentialsReference]:
    ref_id = json.get("id")
    if not isinstance(ref_id, str):
        return DtlError("A credentials reference needs an 'id' field")
    else:
        ref_id = UUID(ref_id)

    ref_type = json.get("type")
    if not isinstance(ref_type, str):
        return DtlError("A credentials reference needs a '_type' field")
    else:
        ref_type = CredentialType.credential_type_from_str(ref_type)

    name = json.get("name")
    return CredentialsReference(ref_id, name, ref_type)


class Credentials(abc.ABC):
    type_field = "type"

    def __init__(self, definition_type: CredentialType):
        self.type = definition_type

    def _base_payload(self) -> dict:
        return dict([(Credentials.type_field, self.type.value)])

    def __eq__(self, other: "Credentials"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """


class S3CredentialType(SerializableStringEnum):
    BasicS3 = "BasicS3"
    RoleS3 = "RoleS3"
    TemporaryS3 = "TemporaryS3"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("S3 Credential Type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, "S3CredentialType"]:
        return SerializableStringEnum.from_str(S3CredentialType)(json)


class S3Credentials(Credentials):
    type_str = CredentialType.S3

    def __init__(
        self,
        region: str,
        connection_bucket: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        endpoint: Optional[str] = None,
        encryption: Optional[Encryption] = None,
        role_arn: Optional[str] = None,
        session_timeout_in_hours: Optional[int] = None,
    ):
        Credentials.__init__(self, S3Credentials.type_str)
        self.region = region
        is_client_id_set = client_id is not None
        is_client_secret_set = client_secret is not None
        if is_client_id_set != is_client_secret_set:
            raise ValueError("`client_id` and `client_secret` must both be set, or neither can be set")
        is_role_arn_set = role_arn is not None
        is_session_timeout_in_hours_set = session_timeout_in_hours is not None
        if is_session_timeout_in_hours_set and (not is_role_arn_set):
            raise ValueError("`role_arn` must be set if `session_timeout_in_hours` specifies a value")
        if is_role_arn_set and (is_client_id_set or is_client_secret_set):
            raise ValueError(
                "Either `client_id` and `client_secret` must be set or `role_arn` be set, both can't be set"
            )
        if is_client_id_set and is_client_secret_set:
            self.client_id = client_id
            self.client_secret = client_secret
            self.credential_type = S3CredentialType.BasicS3
        if is_role_arn_set:
            self.role_arn = role_arn
            if is_session_timeout_in_hours_set:
                self.session_timeout_in_hours = session_timeout_in_hours
            self.credential_type = S3CredentialType.RoleS3
        self.connection_bucket = connection_bucket
        self.endpoint = endpoint
        self.encryption = encryption

    def __repr__(self):
        repr_str = f"{self.__class__.__name__}(region= {self.region!r}, connection_bucket= {self.connection_bucket!r}, "
        if hasattr(self, "client_id") and hasattr(self, "client_secret"):
            repr_str = repr_str + f"client_id= {self.client_id!r}, client_secret= ****, "
        repr_str = repr_str + f"endpoint= {self.endpoint!r}"
        if hasattr(self, "role_arn"):
            repr_str = repr_str + f", role_arn= {self.role_arn!r}"
            if hasattr(self, "session_timeout_in_hours"):
                repr_str = repr_str + f", session_timeout_in_hours= {self.session_timeout_in_hours!r}"
        repr_str = repr_str + ") "
        return repr_str

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        if self.connection_bucket is not None:
            base["connectionBucket"] = self.connection_bucket
        if hasattr(self, "client_id") and hasattr(self, "client_secret"):
            base["clientId"] = self.client_id
            base["clientSecret"] = self.client_secret
            base["credentialType"] = self.credential_type.value
        if self.endpoint is not None:
            base["endpoint"] = self.endpoint
        if self.encryption is not None:
            base["encryption"] = self.encryption._as_payload()
        if hasattr(self, "role_arn"):
            base["roleArn"] = self.role_arn
            if hasattr(self, "session_timeout_in_hours"):
                base["sessionDurationInSec"] = self._validate_and_convert_to_seconds_session_timeout(
                    self.session_timeout_in_hours
                )
            base["credentialType"] = self.credential_type.value
        return base

    @staticmethod
    def _validate_and_convert_to_seconds_session_timeout(
        session_timeout_in_hours,
    ) -> Union[int, DtlError]:
        if not 1 <= session_timeout_in_hours <= 12:
            raise DtlError("The `session_timeout_in_hours` should be between 1 and 12 hours")
        else:
            return session_timeout_in_hours * 3600


class S3ACredentials(S3Credentials):
    type_str = CredentialType.Hadoop

    def __init__(self, region: str, client_id: str, client_secret: str, endpoint: str):
        S3Credentials.__init__(self, region, client_id, client_secret, endpoint)
        self.type = S3ACredentials.type_str


class AmazonVendorCentral(Credentials):
    type_str = CredentialType.AmazonVendorCentral

    def __init__(self, client_id: str, client_secret: str, mws_auth_token: str, seller_id: str):
        Credentials.__init__(self, AmazonVendorCentral.type_str)
        self.client_id = client_id
        self.client_secret = client_secret
        self.mws_auth_token = mws_auth_token
        self.seller_id = seller_id

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(client_id= ***, client_secret= ***, "
            f"mws_auth_token= ***, seller_id= {self.seller_id})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        base["mwsAuthToken"] = self.mws_auth_token
        base["sellerId"] = self.seller_id
        return base


class AzureBlobType(SerializableStringEnum):
    Block = "Block"
    Append = "Append"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, "AzureBlobType"]:
        return SerializableStringEnum.from_str(AzureBlobType)(json)


class AzureEndpointProtocol(SerializableStringEnum):
    Https = "Https"
    Http = "Http"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, "AzureBlobType"]:
        return SerializableStringEnum.from_str(AzureEndpointProtocol)(json)


class AzureCredentials(Credentials):
    type_str = CredentialType.Azure

    def __init__(
        self,
        account_name: str,
        account_key: str,
        endpoint_protocol: AzureEndpointProtocol,
        blob_type: AzureBlobType,
    ):
        Credentials.__init__(self, AzureCredentials.type_str)
        self.account_name = account_name
        self.account_key = account_key
        self.endpoint_protocol = endpoint_protocol
        self.blob_type = blob_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(account_name= {self.account_name!r},"
            f" endpoint_protocol= {self.endpoint_protocol!r}, blob_type= {self.blob_type!r}, account_key= ****)"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["accountName"] = self.account_name
        base["accountKey"] = self.account_key
        base["endpointProtocol"] = self.endpoint_protocol.value
        base["blobType"] = self.blob_type.value
        return base


class HadoopAzureCredentials(Credentials):
    type_str = CredentialType.HadoopAzure

    def __init__(self, account_name: str, account_key: str):
        Credentials.__init__(self, HadoopAzureCredentials.type_str)
        self.account_name = account_name
        self.account_key = account_key

    def __repr__(self):
        return f"{self.__class__.__name__}(account_name= {self.account_name!r}," f" account_key= ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["accountName"] = self.account_name
        base["accountKey"] = self.account_key
        return base


class AzureSASCredentials(Credentials):
    type_str = CredentialType.Azure

    def __init__(self, sas_token: str, endpoint: str):
        Credentials.__init__(self, AzureSASCredentials.type_str)
        self.sas_token = sas_token
        self.endpoint = endpoint

    def __repr__(self):
        return f"{self.__class__.__name__}(sas_token= ****," f" endpoint= {self.endpoint!r})"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["sas_token"] = self.sas_token
        base["endpoint"] = self.endpoint
        return base


class GCS(Credentials):
    type_str = CredentialType.GCS

    def __init__(self, client_email: str, private_key: str):
        Credentials.__init__(self, GCS.type_str)
        self.client_email = client_email
        self.private_key = private_key

    def __repr__(self):
        return f"{self.__class__.__name__}(client_email= {self.client_email!r}, private_key= ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["clientEmail"] = self.client_email
        base["privateKey"] = self.private_key
        return base


class JdbcCredentials(Credentials):
    type_str = CredentialType.JDBC

    def __init__(self, url: str, user: str, password: str):
        Credentials.__init__(self, JdbcCredentials.type_str)
        self.url = url
        self.user = user
        self.password = password

    def __repr__(self):
        return f"{self.__class__.__name__}(url= {self.url!r}, user= {self.user!r}, password= ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["user"] = self.user
        base["password"] = self.password
        return base


class MongoCredentials(Credentials):
    type_str = CredentialType.Mongo

    def __init__(self, url: str, database: str, user: str, password: str):
        Credentials.__init__(self, MongoCredentials.type_str)
        self.url = url
        self.database = database
        self.user = user
        self.password = password

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(url= {self.url!r}, database= {self.database!r}, user= {self.user!r}, "
            f"password= ****) "
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["database"] = self.database
        base["user"] = self.user
        base["password"] = self.password
        return base


class SocrataCredentials(Credentials):
    type_str = CredentialType.Socrata

    def __init__(self, domain: str, token: str):
        Credentials.__init__(self, SocrataCredentials.type_str)
        self.domain = domain
        self.token = token

    def __repr__(self):
        return f"{self.__class__.__name__}(domain= {self.domain!r}, token= ****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["domain"] = self.domain
        base["token"] = self.token
        return base


class KinesisCredentials(Credentials):
    type_str = CredentialType.Kinesis

    def __init__(self, region: str, client_id: str, client_secret: str):
        Credentials.__init__(self, KinesisCredentials.type_str)
        self.region = region
        self.client_id = client_id
        self.client_secret = client_secret

    def __repr__(self):
        return f"{self.__class__.__name__}(region= {self.region!r}, client_id= {self.client_id!r}, client_secret= ****)"

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        return base


class KerberosCredentials(Credentials):
    type_str = CredentialType.Kerberos

    def __init__(self, principal: str, password: str, with_ssl: bool = True):
        Credentials.__init__(self, KerberosCredentials.type_str)
        self.principal = principal
        self.password = password
        self.with_ssl = with_ssl

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(principal= {self.principal!r}, "
            f"password= ****, "
            f"with_ssl= {self.with_ssl})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["principal"] = self.principal
        base["password"] = self.password
        base["withSsl"] = self.with_ssl
        base["type"] = "Kerberos"
        return base


class JaasString(Credentials):
    type_str = CredentialType.JaasString

    def __init__(self, sasl_jaas_config: str, with_ssl: bool = True):
        Credentials.__init__(self, KerberosCredentials.type_str)
        self.sasl_jaas_config = sasl_jaas_config
        self.with_ssl = with_ssl

    def __repr__(self):
        return (
            f"{self.__class__.__name__}" f"(saslJaasConfig= {self.sasl_jaas_config!r}, " f"with_ssl= {self.with_ssl})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["saslJaasConfig"] = self.sasl_jaas_config
        base["withSsl"] = self.with_ssl
        base["type"] = "JaasString"
        return base


class KafkaCredentials(Credentials):
    type_str = CredentialType.Kafka

    def __init__(
        self,
        bootstrap_server: str,
        registry_url: Optional[str] = None,
        auth: Union[KerberosCredentials, JaasString, None] = None,
        change_data_source: Optional[ChangeDataSource] = None,
    ):
        Credentials.__init__(self, KafkaCredentials.type_str)
        self.bootstrap_server = bootstrap_server
        self.registry_url = registry_url
        self.auth = auth
        self.change_data_source = change_data_source

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(bootstrap_server= {self.bootstrap_server!r}, "
            f"registry_url= {self.registry_url}, "
            f"auth= {self.auth}, "
            f"change_data_source= {self.change_data_source})"
        )

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["bootstrapServer"] = self.bootstrap_server
        if self.registry_url is not None:
            base["registryUrl"] = self.registry_url
        if isinstance(self.auth, KerberosCredentials):
            base["auth"] = KerberosCredentials._as_payload(self.auth)
        if isinstance(self.auth, JaasString):
            base["auth"] = JaasString._as_payload(self.auth)
        if self.change_data_source is not None:
            base["changeDataSource"] = self.change_data_source
        return base


class ResourceType(SerializableStringEnum):
    """
    Class that handles all ResourceTypes
    """

    File = "File"
    Directory = "Directory"
    Unknown = "Unknown"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("Resource type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, "ResourceType"]:
        return SerializableStringEnum.from_str(ResourceType)(string)


class SFTPCredentials(Credentials):
    type_str = CredentialType.SFTP

    def __init__(
        self,
        address: str,
        port: int,
        user_name: str,
        private_key: str,
        encryption_key: str = "",
        target_resource_path: str = "",
        target_resource_type: Optional[ResourceType] = None,
    ):
        """

        :param address: Hostname or IP address of the SFTP server
        :param port: Port number
        :param user_name User name to be used to connect to the remote machine
        :param private_key: Private Key to be used to connect to the remote machine. The key has to be in PEM format and
          not OpenSSH format
        :param encryption_key Password for the private key itself, Defaults to empty string for no password
        :param target_resource_path TODO
        :param target_resource_type TODO
        """
        Credentials.__init__(self, SFTPCredentials.type_str)
        self.address = address
        self.port = port
        self.user_name = user_name
        self.private_key = private_key
        self.encryption_key = encryption_key
        self.target_resource_path = target_resource_path
        self.target_resource_type = target_resource_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(address= {self.address}, port= {self.port}, user_name= {self.user_name}, "
            f"private_key= ****, encryption_key= ****,"
            f"target_resource_path= '{self.target_resource_path}', target_resource_type= {self.target_resource_type!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["address"] = self.address
        base["port"] = self.port
        base["userName"] = self.user_name
        base["privateKey"] = self.private_key
        base["encryptionKey"] = self.encryption_key
        base["targetResourcePath"] = self.target_resource_path
        if self.target_resource_type is not None:
            base["targetResourceType"] = self.target_resource_type.value
        return base


class HTTPCredentials(Credentials):
    type_str = CredentialType.Http

    def __init__(self, base_url: str, headers: Optional[Dict[str, List[str]]] = None):
        """
        :param base_url: given base url to fetch/send the content
        :param headers: headers to be added to the request (Accept, X-App-Token, etc.). it's optional
        """
        Credentials.__init__(self, HTTPCredentials.type_str)
        self.base_url = base_url
        self.headers = headers

    def __repr__(self):
        return f"{self.__class__.__name__}(base_url= {self.base_url}, headers= {self.headers!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["baseUrl"] = self.base_url
        if self.headers is not None:
            base["headers"] = self.headers
        else:
            base["headers"] = {}
        return base


class ElasticSearchCredentials(Credentials):
    type_str = CredentialType.ElasticSearch

    def __init__(self, url: str, username: str, password: str):
        Credentials.__init__(self, ElasticSearchCredentials.type_str)
        self.url = url
        self.username = username
        self.password = password

    def __repr__(self):
        return f"{self.__class__.__name__}(url={self.url!r},username={self.username!r}," f"password=****)"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["username"] = self.username
        base["password"] = self.password
        return base


class WebhookCredentials(Credentials):
    type_str = CredentialType.Webhook

    def __init__(
        self,
        file_format: FileFormat,
        identifier: Optional[str] = None,
        webhook_url: Optional[str] = None,
        buffer: int = 0,
    ):
        Credentials.__init__(self, WebhookCredentials.type_str)
        self.file_format = file_format
        self.buffer = buffer
        self.identifier = identifier
        self.webhook_url = webhook_url

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(file_format={self.file_format!r},identifier={self.identifier!r},"
            f"webhook_url={self.webhook_url!r},buffer={self.buffer!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["fileFormat"] = self.file_format.value
        base["buffer"] = self.buffer
        if self.webhook_url is not None:
            base["webhookUrl"] = self.webhook_uri
        if self.identifier is not None:
            base["identifier"] = self.identifier
        return base


class BoxCredentials(Credentials):
    type_str = CredentialType.Box

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        public_key_id: str,
        private_key_password: str,
        private_key: str,
        enterprise_id: str = "737833",
    ):
        """
        :param client_id: client id for a box app
        :param client_secret: client secret for a box app
        :param public_key_id:  public key id  belonging to a box app
        :param private_key_password: passphrase for the given private key belonging to a box app
        :param private_key: private key belonging to a box app or service account
        :param enterprise_id: unique id representing a given box enterprise account. 737833 is the id
        for Nike

        All these parameters can be found in a json file that gets created when trying to
        generate a public/private keypair for a box app.
        """
        Credentials.__init__(self, BoxCredentials.type_str)
        self.client_id = client_id
        self.client_secret = client_secret
        self.public_key_id = public_key_id
        self.private_key_password = private_key_password
        self.private_key = private_key
        self.enterprise_id = enterprise_id

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(client_id={self.client_id!r},client_secret=***,"
            f"public_key_id=***,private_key_password=***,private_key=****,enterprise_id={self.enterprise_id!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        base["publicKeyId"] = self.public_key_id
        base["privateKey"] = self.private_key
        base["privateKeyPassword"] = self.private_key_password
        base["enterpriseId"] = self.enterprise_id
        return base



import os
from typing import Optional
from urllib.parse import urlparse

from datalogue.errors import DtlError


class DtlCredentials:
    """
    Submit credentials to log in to Datalogue
    Note: this is not an Okta login. If using Okta, please generate a token via the web app and use DtlTokenCredentials
    if verify_certificate is set, this variable will be used otherwise we look at the environment variable
    `DTL_SSL_VERIFY_CERT` if none is set then we use True as a default

    :param username: username to be used to login on the platform
    :param password: password to be used to login on the platform
    :param uri: root url where the system lives ie: https://app.dtl.nike.com/api
    :param verify_certificate: (Optional) verify the certificate when connecting to remote, no value means true.
    :param authentication_name: can be used to specify login via a configured alternate authentication method,
    e.g. LDAP. If None, will use the default email/password login
    """

    def __init__(
        self,
        username: str,
        password: str,
        uri: str = "https://app.dtl.nike.com/api",
        verify_certificate: Optional[bool] = None,
        authentication_name: Optional[str] = None,
    ):
        self.username = username.strip()
        self.password = password.strip()

        if verify_certificate is None:
            verify_certificate = _get_ssl_verify_env()

        self.verify_certificate = verify_certificate

        self.authentication_name = authentication_name
        uri = uri.strip()

        if self.validate_url(uri) is not True:
            raise DtlError("The URL you provided is invalid")

        if not uri.endswith("/api"):
            raise DtlError("The URL you provided doesn't end with '/api' it is most likely invalid")

        self.uri = uri

    def validate_url(self, uri: str) -> bool:
        parsed_url = urlparse(uri)
        return all([parsed_url.scheme, parsed_url.netloc])

    def __repr__(self):
        res = f"{self.__class__.__name__}(username: {self.username!r}, password: ****, uri: {self.uri!r})"
        if self.verify_certificate is None:
            return res
        else:
            return res[:-1] + f", verify_certificate: {self.verify_certificate!r})"


def _get_ssl_verify_env() -> bool:
    verify_certificate = os.environ.get("DTL_SSL_VERIFY_CERT")
    return verify_certificate is None or verify_certificate.lower() != "false"



from typing import List, Union, NamedTuple, Literal
from uuid import UUID
from datetime import datetime, timedelta
from datalogue.dtl_utils import (
    _parse_list,
    SerializableStringEnum,
)
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models import ComparisonOperator


class Metric(SerializableStringEnum):
    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("metric", s)

    @staticmethod
    def _from_payload(json: str) -> Union[DtlError, "Metric"]:
        return SerializableStringEnum.from_str(Metric)(json)

    TOTAL_RECORD_COUNT = "TotalRecordCount"
    REJECTED_RECORD_COUNT = "RejectedRecordCount"


class Comparison(NamedTuple):
    """
    Comparison(TOTAL_RECORD_COUNT, LT, 100) means
    an alert will be triggered if TotalRecordCount < 100
    """
    metric: Metric
    operator: ComparisonOperator
    value: float

    def _as_payload(self):
        return {
            "type": "Comparison",
            "metric": self.metric.value,
            "operator": self.operator.value,
            "value": self.value,
        }


class JobDuration(NamedTuple):
    """
    JobDuration(timedelta) means an alert will be
    triggered if the job runtime exceeds the specified timedelta
    """
    duration: timedelta

    def _as_payload(self):
        return {
            "type": "JobDuration",
            "durationSeconds": self.duration.total_seconds()
        }

class EmailChannel(NamedTuple):
    email: List[str]

    def _as_payload(self):
        return {
            "type": "Email",
            "email": self.email
        }


class PagerDuty(NamedTuple):
    integration_key: str

    def _as_payload(self):
        return {
            "type": "PagerDuty",
            "integrationKey": self.integration_key
        }


class SlackChannel(NamedTuple):
    webhook_url: str

    def _as_payload(self):
        return {
            "type": "Slack",
            "webhook": self.webhook_url
        }


AlertCondition = Union[Literal["JobFailure", "JobCancellation", "NoInput"], Comparison, JobDuration]
AlertChannel = Union[EmailChannel, PagerDuty, SlackChannel]


def _alert_channel_from_payload(json: dict) -> Union[DtlError, AlertChannel]:
    t = json.get("type")
    if t == "Email":
        email = json.get("email")
        if not isinstance(email, str):
            return DtlError("Email in EmailChannel should be a string")
        return EmailChannel(email)
    elif t == "PagerDuty":
        integration_key = json.get("integrationKey")
        if not isinstance(integration_key, str):
            return DtlError("Escalation policy in PagerDuty should be a string")
        return PagerDuty(
            integration_key=integration_key
        )
    elif t == "Slack":
        webhook_url = json.get("webhook")
        if not isinstance(webhook_url, str):
            return DtlError("Channel name in SlackChannel should be a string")
        return SlackChannel(webhook_url)


def _alert_condition_from_payload(json: dict) -> Union[DtlError, AlertCondition]:
    t = json.get("type")
    if t in ["JobFailure", "JobCancellation", "NoInput"]:
        return t
    elif t == "Comparison":
        metric = Metric._from_payload(json.get("metric"))
        if isinstance(metric, DtlError):
            return metric
        operator = ComparisonOperator._from_payload(json.get("operator"))
        if isinstance(operator, DtlError):
            return operator
        value = json.get("value")
        if not isinstance(value, float):
            return DtlError("Value in comparison should be a float")
        return Comparison(
            metric=metric, operator=operator, value=value
        )
    elif t == "JobDuration":
        duration_seconds = json.get("durationSeconds")
        if not isinstance(duration_seconds, int):
            return DtlError("The value of durationSeconds should be an int")
        return JobDuration(
            duration=timedelta(seconds=duration_seconds)
        )
    else:
        return DtlError("Unknown type for AlertCondition")


class Alert(NamedTuple):
    """
    :param id: ID of the alert, leave None when creating
    :param name: Name of the alert
    :param pipeline_ids: IDs of pipelines to check
    :param triggers: any of the alert conditions will trigger the alert
    :param channels: ways to be notified
    :param activated: no alerts will be sent if the alert is deactivated
    """
    id: UUID
    name: str
    pipeline_ids: List[UUID]
    triggers: List[AlertCondition]
    channels: List[AlertChannel]
    activated: bool = False

    def _as_payload(self) -> dict:
        return {
            "id": str(self.id),
            "name": self.name,
            "pipelineIds": list(map(lambda i: str(i), self.pipeline_ids)),
            "triggers": list(map(lambda t: {"type": t} if isinstance(t, str) else t._as_payload(), self.triggers)),
            "channels": list(map(lambda c: c._as_payload(), self.channels)),
            "activated": self.activated,
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Alert"]:
        alert_id = json.get("id")
        if alert_id is None:
            return DtlError("Alert object should have a 'id' property")
        alert_id = UUID(alert_id)

        name = json.get("name")
        if name is None:
            return DtlError("Alert object should have a 'name' property")

        pipeline_ids = json.get("pipelineIds")
        if pipeline_ids is None:
            return DtlError("Alert object should have a 'pipeline_ids' property")

        triggers = json.get("triggers")
        if triggers is None:
            return DtlError("Alert object should have a 'triggers' property")
        triggers = _parse_list(_alert_condition_from_payload)(triggers)
        channels = json.get("channels")
        if channels is None:
            return DtlError("Alert object should have a 'channels' property")
        channels = _parse_list(_alert_channel_from_payload)(channels)
        activated = json.get("activated")
        if not isinstance(activated, bool):
            return DtlError("Alert object should have a boolean 'activated' property")
        return Alert(
            id=alert_id,
            name=name,
            pipeline_ids=pipeline_ids,
            triggers=triggers,
            channels=channels,
            activated=activated,
        )



from typing import List, Union, Optional
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.models.classifier import (
    ClassificationMethod,
    _classification_method_from_payload,
)
from datalogue.dtl_utils import _parse_list
from uuid import UUID


# Kept because classify_fields is still depending on it
class Classifier(object):
    def __init__(
        self,
        classification_methods: List[ClassificationMethod],
        default_class_id: Optional[Union[UUID, str]] = None,
    ):
        """
        Build a local Classifier

        :param classification_methods: ordered list of Classification Methods
        :param default_class_id: allows the user to specify the class to be defaulted to, if all classification methods
            fail. Default value is `Unknown`. (currently part of replace_class, see below)
        """
        self.classification_methods = classification_methods
        self.default_class_id = default_class_id

    def __repr__(self):
        return (
            "Classifier("
            f"default_class= {self.default_class_id} "
            f"classification_methods= \n{self.classification_methods})"
        )

    def __eq__(self, other: "Classifier"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        base = {
            "classificationMethods": list(
                map(lambda m: m._as_payload(), self.classification_methods)
            ),
            "defaultClass": self.default_class_id,
        }
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Classifier"]:
        classification_methods = json.get("classificationMethods")
        if classification_methods is None:
            return _property_not_found("classificationMethods", json)
        else:
            classification_methods = _parse_list(_classification_method_from_payload)(
                classification_methods
            )
            if isinstance(classification_methods, DtlError):
                return classification_methods

        default_class = json.get("default_class")

        return Classifier(classification_methods, default_class)


class Classify(Transformation):
    """
    Apply classes to all data streaming through a pipeline. Classification is based on the Classifier supplied.
    """

    type_str = "Classify"

    def __init__(
        self,
        classifier_id: Union[str, UUID],
        fields_to_target: Optional[List[List[str]]] = None,
        add_class_fields=False,
        add_score_fields=False,
    ):
        """
        Unless specified with parameters below, this transformation does not alter the data schema; classes are
        represented in the backend only.

        :param classifier_id: id of the classifier to use for classification
        :param fields_to_target: specifies which fields to target for classification; remainder are not classified by
         this transformation
        :param add_class_fields: if True, adds a field containing class name for every classified field
        :param add_score_fields: if True, adds a field containing class score for every classified field
        :param use_uuid: When enabled, pipeline will be able to compare by classId instead of class names. requirements:
        1. add_class_fields and add_score_fields must be false
        2. Currently works for Obfuscate transformation

        """

        Transformation.__init__(self, Classify.type_str)
        self.classifier_id = classifier_id
        self.fields_to_target = fields_to_target
        self.add_class_fields = add_class_fields
        self.add_score_fields = add_score_fields

    def __eq__(self, other: "Classify"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return (
            "Classify("
            f"classifier_id= {self.classifier_id}, "
            f"fields_to_target= {self.fields_to_target}, "
            f"add_class_fields= {self.add_class_fields}, "
            f"add_score_fields= {self.add_score_fields})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["classifier"] = str(self.classifier_id)
        base["fieldsToTarget"] = self.fields_to_target
        base["addClassFields"] = self.add_class_fields
        base["addScoreFields"] = self.add_score_fields
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Classify"]:
        classifier_id = json.get("classifier")
        if classifier_id is None:
            return _property_not_found("classifier", json)

        fields_to_target = json.get("fieldsToTarget")

        add_class_fields = json.get("addClassFields")
        if add_class_fields is None:
            add_class_fields = False

        add_score_fields = json.get("addScoreFields")
        if add_score_fields is None:
            add_score_fields = False

        return Classify(
            classifier_id, fields_to_target, add_class_fields, add_score_fields
        )
